/* Copyright (c) 2000 John Stracke <fanorona@thibault.org>
 #
 * This file is part of Fanorona.
 #
 * Fanorona is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version. 
 #
 * Fanorona is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details. 
 #
 * You should have received a copy of the GNU General Public License along
 * with program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <errno.h>
#include <fanorona/player.h>

typedef struct _FanoronaPlayerLookaheadNode FanoronaPlayerLookaheadNode;

typedef struct {
  int byMe;
  struct {
    int col,row;
  } from,to;
  int byApproach;
} FanoronaPlayerMove;

struct _FanoronaPlayerLookaheadNode {
  /* Move being considered */
  FanoronaPlayerMove move;

  /* The number of pieces each player has if this move is taken. */
  struct {
    int mine,other;
  } counts;

  int moveNumber;

  FanoronaPlayerLookaheadNode* nextSib;
  FanoronaPlayerLookaheadNode* prevSib;
  FanoronaPlayerLookaheadNode* firstChild;
};

struct _FanoronaPlayer {
  int isPlayer1,depth,curDepth;
  FanoronaBoard curBoard;
  FanoronaPlayerLookaheadNode* lookahead;
};

static void freeLookaheadTree(FanoronaPlayerLookaheadNode* tree)
{
  /*printf("freeLookaheadTree(%p): moveNumber %d\n",
         tree,tree->moveNumber);*/
  fflush(stdout); fflush(stderr);

  if (!tree)
    return;

  if (tree->nextSib)
    freeLookaheadTree(tree->nextSib);
  if (tree->firstChild)
    freeLookaheadTree(tree->firstChild);

  free(tree);
}

static int movesEqual(const FanoronaPlayerMove* move1,
                      const FanoronaPlayerMove* move2)
{
  if (!(move1 && move2))
    return 0;

  return (((move1->byMe)==(move2->byMe))
          && ((move1->from.col)==(move2->from.col))
          && ((move1->from.row)==(move2->from.row))
          && ((move1->to.col)==(move2->to.col))
          && ((move1->to.row)==(move2->to.row))
          && ((move1->byApproach)==(move2->byApproach))
          );
}

static void indent(int cols)
{
  while (cols>0)
    {
      printf(" ");
      cols--;
    }
}

static void _dumpMove(const FanoronaPlayerMove* move,
                      int indents)
{
  indent(indents);
  printf("<move byMe='%s' from='(%d,%d)' to='(%d,%d)' by='%s'/>\n",
         (move->byMe ? "yes" : "no"),
         move->from.col,
         move->from.row,
         move->to.col,
         move->to.row,
         (move->byApproach ? "approach" : "withdrawal"));
  fflush(stdout); fflush(stderr);
}

static void dumpMove(const FanoronaPlayerMove* move)
{
  _dumpMove(move,0);
}

static void _dumpTree(const FanoronaPlayerLookaheadNode* tree,
                      int indents)
{
  const FanoronaPlayerLookaheadNode* cur;

  for (cur=tree; cur; cur=cur->nextSib)
    {
      indent(indents);
      printf("<node moveNumber='%d' addr='%p' nextSib='%p'>\n",
             cur->moveNumber,
             cur,
             cur->nextSib);
      _dumpMove(&(cur->move),indents+1);
      indent(indents+1);
      printf("<counts mine='%d' other='%d'>\n",
             cur->counts.mine,cur->counts.other);
      if (cur->firstChild)
        _dumpTree(cur->firstChild,indents+1);
      indent(indents);
      printf("</node>\n");
    }
  fflush(stdout); fflush(stderr);
}

static void dumpTree(const FanoronaPlayerLookaheadNode* tree)
{
  if (!tree)
    {
      printf("null tree\n");
      return;
    }

  printf("<tree>\n");
  _dumpTree(tree,1);
  printf("</tree>\n");
}

static int copyAndMove(FanoronaBoard* dest,
                       const FanoronaBoard* src,
                       const FanoronaPlayerMove* move)
{
  if (fanoronaBoardCopyAndMove(dest,
                               src,
                               move->from.col,
                               move->from.row,
                               move->to.col,
                               move->to.row,
                               move->byApproach)
      )
    {
      printf("copyAndMove() failed: move is ");
      dumpMove(move);
      return -1;
    }

  return 0;
}

static FanoronaPlayerLookaheadNode*
pruneLookaheadTree(FanoronaPlayer* player,
                   const FanoronaPlayerMove* moveMade)
{
  fflush(stdout); fflush(stderr);
  if (!player)
    {
      fprintf(stderr,"pruneLookaheadTree(): player is null\n");
      errno=EFAULT;
      return 0;
    }

  if (!(player->lookahead))
    {
      fprintf(stderr,"pruneLookaheadTree(): nothing to prune\n");
      return 0;
    }

  if (!moveMade)
    {
      fprintf(stderr,"pruneLookaheadTree(): moveMade is null\n");
      errno=EFAULT;
      return 0;
    }

  /*dumpTree(player->lookahead);*/

  {
    FanoronaPlayerLookaheadNode* cur;
    FanoronaPlayerLookaheadNode* remaining;

    for (cur=player->lookahead; cur; cur=cur->nextSib)
      {
        if (movesEqual(moveMade,&(cur->move)))
          break;
      }

    if (!cur)
      {
        fprintf(stderr,"pruneLookaheadTree(): moveMade not found\n");
        errno=ENOENT;
        return 0;
      }

    remaining=cur->firstChild;
    cur->firstChild=0;

    /*printf("pruneLookaheadTree(): cur==%p, remaining==%p\n",
           cur,remaining);*/

    fflush(stdout); fflush(stderr);
    freeLookaheadTree(player->lookahead);
    fflush(stdout); fflush(stderr);

    player->lookahead=remaining;
    player->curDepth--;

    /*dumpTree(player->lookahead);*/

    return remaining;
  }
}

static int applyMove(FanoronaPlayer* player,
                     const FanoronaPlayerMove* move)
{
  pruneLookaheadTree(player,move);
  if (fanoronaBoardMove(&(player->curBoard),
                        move->from.col,
                        move->from.row,
                        move->to.col,
                        move->to.row,
                        move->byApproach)<0)
    {
      int tmp=errno;
      fprintf(stderr,"applyMove(): could not do move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  return 0;
}

int fanoronaPlayerInit(FanoronaPlayer* player,
                       int isPlayer1,
                       int depth)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerInit(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (fanoronaBoardSetup(&(player->curBoard))<0)
    {
      fprintf(stderr,"fanoronaPlayerInit(): can't set up board\n");
      errno=EFAULT;
      return -1;
    }

  player->isPlayer1=isPlayer1;
  player->depth=depth;
  player->curDepth=0;
  player->lookahead=0;

  return 0;
}

FanoronaPlayer* fanoronaPlayerNew(int isPlayer1,
                                  int depth)
{
  FanoronaPlayer* res=(FanoronaPlayer*)malloc(sizeof(FanoronaPlayer));
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerNew(): can't malloc player: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  if (fanoronaPlayerInit(res,isPlayer1,depth)<0)
    {
      int tmp=errno;
      free(res);
      fprintf(stderr,"fanoronaPlayerNew(): can't init player: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  return res;
}

void fanoronaPlayerFree(FanoronaPlayer* player)
{
  if (!player)
    return;

  if (player->lookahead)
    freeLookaheadTree(player->lookahead);
  free(player);
}

const FanoronaBoard* fanoronaPlayerGetBoard(const FanoronaPlayer* player)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerGetBoard(): player is null\n");
      errno=EFAULT;
      return 0;
    }
  
  return &(player->curBoard);
}

int fanoronaPlayerIsPlayer1(FanoronaPlayer* player)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerGetBoard(): player is null\n");
      errno=EFAULT;
      return 0;
    }
  
  return player->isPlayer1;
}

static int deepenLookaheadBy1Leaf(FanoronaPlayer* player,
                                  const FanoronaBoard* board,
                                  FanoronaPlayerLookaheadNode** leaf,
                                  int isMyMove,
                                  int moveNumber)
{
  FanoronaBoardCellState myState,otherState,movingState,nonMovingState;
  FanoronaPlayerLookaheadNode* lastChild=0;

  if (!player)
    {
      fprintf(stderr,"deepenLookaheadBy1Leaf(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (*leaf)
    {
      fprintf(stderr,"deepenLookaheadBy1Leaf(): *leaf is not null\n");
      errno=EFAULT;
      return -1;
    }

  if (player->isPlayer1)
    {
      myState=fanoronaBoardCellStatePlayer1;
      otherState=fanoronaBoardCellStatePlayer2;
    }
  else
    {
      myState=fanoronaBoardCellStatePlayer2;
      otherState=fanoronaBoardCellStatePlayer1;
    }

  if (isMyMove)
    {
      movingState=myState;
      nonMovingState=otherState;
    }
  else
    {
      movingState=otherState;
      nonMovingState=myState;
    }

  {
    int colFrom,rowFrom;
    for (colFrom=0; colFrom<FANORONA_NUM_COLS; colFrom++)
      for (rowFrom=0; rowFrom<FANORONA_NUM_ROWS; rowFrom++)
        {
          int colTo,rowTo;
          if (fanoronaBoardGetCell(board,colFrom,rowFrom)!=movingState)
            continue;

          for (colTo=colFrom-1; colTo<=(colFrom+1); colTo++)
            for (rowTo=rowFrom-1; rowTo<=(rowFrom+1); rowTo++)
              {
                int byApproach,canApproach,canWithdraw,dc,dr;

                if ((colTo==colFrom) && (rowTo==rowFrom))
                  continue;

                if (!fanoronaBoardIsLegalPos(colTo,rowTo))
                  continue;

                if (fanoronaBoardGetCell(board,colTo,rowTo)
                    !=fanoronaBoardCellStateEmpty
                    )
                  continue;

                if (!fanoronaBoardTestAdjacent(colFrom,rowFrom,colTo,rowTo))
                  continue;

                dc=colTo-colFrom;
                dr=rowTo-rowFrom;

                canApproach
                  =(   fanoronaBoardIsLegalPos(colTo+dc,rowTo+dr)
                    && (fanoronaBoardGetCell(board,colTo+dc,rowTo+dr)
                        ==nonMovingState)
                       );
                canWithdraw
                  =(   fanoronaBoardIsLegalPos(colTo-dc,rowTo-dr)
                    && (fanoronaBoardGetCell(board,colTo-dc,rowTo-dr)
                        ==nonMovingState)
                       );

                if (!(canApproach || canWithdraw))
                  canApproach=1;

                for (byApproach=0; byApproach<2; byApproach++)
                  {
                    FanoronaBoard potentialBoard;
                    FanoronaPlayerLookaheadNode* newChild=0;

                    if (byApproach)
                      {
                        if (!canApproach)
                          continue;
                      }
                    else
                      {
                        if (!canWithdraw)
                          continue;
                      }

                    newChild=(FanoronaPlayerLookaheadNode*)
                      malloc(sizeof(FanoronaPlayerLookaheadNode));
                    if (!newChild)
                      {
                        int tmp=errno;
                        fprintf(stderr,
                                "deepenLookaheadBy1Leaf(): can't malloc newChild: %s\n",
                                strerror(tmp));
                        tmp=errno;
                        return -1;
                      }
                    /*printf("newChild==%p\n",newChild);*/

                    newChild->moveNumber=moveNumber;

                    newChild->move.byMe=isMyMove;
                    newChild->move.from.col=colFrom;
                    newChild->move.from.row=rowFrom;
                    newChild->move.to.col=colTo;
                    newChild->move.to.row=rowTo;
                    newChild->move.byApproach=byApproach;

                    /*printf("Trying move: from (%d,%d) to (%d,%d), by %s\n",
                           colFrom,rowFrom,colTo,rowTo,
                           (byApproach ? "approach" : "withdrawal"));
                           */

                    if (fanoronaBoardCopyAndMove(&potentialBoard,board,
                                                 colFrom,rowFrom,
                                                 colTo,rowTo,
                                                 byApproach)<0)
                      {
                        int tmp=errno;
                        fprintf(stderr,
                                "deepenLookaheadBy1Leaf(): can't apply possible move: %s\n",
                                strerror(tmp));
                        tmp=errno;
                        return -1;
                      }

                    {
                      int numForPlayer1,numForPlayer2;
                      fanoronaBoardCountSides(&potentialBoard,
                                              &numForPlayer1,
                                              &numForPlayer2);
                      if (player->isPlayer1)
                        {
                          newChild->counts.mine=numForPlayer1;
                          newChild->counts.other=numForPlayer2;
                        }
                      else
                        {
                          newChild->counts.mine=numForPlayer2;
                          newChild->counts.other=numForPlayer1;
                        }
                    }

                    /*printf("Inserted move: from (%d,%d) to (%d,%d), by %s\n",
                           newChild->move.from.col,newChild->move.from.row,
                           newChild->move.to.col,newChild->move.to.row,
                           (newChild->move.byApproach ? "approach" : "withdrawal"));
                           */

                    if (lastChild)
                      {
                        lastChild->nextSib=newChild;
                        newChild->prevSib=lastChild;
                      }
                    else
                      {
                        (*leaf)=newChild;
                        newChild->prevSib=0;
                      }
                    newChild->firstChild=0;
                    newChild->nextSib=0;
                    lastChild=newChild;
                  }
              }
        }
  }

  fflush(stdout); fflush(stderr);
  return 0;
}

static int deepenLookaheadBy1(FanoronaPlayer* player,
                              const FanoronaBoard* board,
                              FanoronaPlayerLookaheadNode* tree)
{
  FanoronaBoard newBoard;

  if (!player)
    {
      fprintf(stderr,"deepenLookaheadBy1(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!tree)
    {
      tree=player->lookahead;
      if (!tree)
        {
          if (deepenLookaheadBy1Leaf(player,
                                     &(player->curBoard),
                                     &(player->lookahead),
                                     (player->isPlayer1),
                                     1)
              <0)
            {
              int tmp=errno;
              fprintf(stderr,"deepenLookaheadBy1(): can't deepen player: %s\n",
                      strerror(tmp));
              errno=tmp;
              return -1;
            }

          return 0;
        }
    }

  if (tree->nextSib)
    {
      if (deepenLookaheadBy1(player,board,tree->nextSib)<0)
        {
          int tmp=errno;
          fprintf(stderr,"deepenLookaheadBy1(): can't deepen nextSib: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }
    }
  
  copyAndMove(&newBoard,board,&(tree->move));
  
  if (tree->firstChild)
    {
      if (deepenLookaheadBy1(player,&newBoard,tree->firstChild)<0)
        {
          int tmp=errno;
          fprintf(stderr,"deepenLookaheadBy1(): can't recurse into firstChild: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }
    }
  else
    {
      if (deepenLookaheadBy1Leaf(player,
                                 &newBoard,
                                 &(tree->firstChild),
                                 !(tree->move.byMe),
                                 tree->moveNumber+1)
          <0)
        {
          int tmp=errno;
          fprintf(stderr,"deepenLookaheadBy1(): can't deepen firstChild: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }
    }

  return 0;
}

static int deepenLookahead(FanoronaPlayer* player)
{
  if (!player)
    {
      fprintf(stderr,"deepenLookahead(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  while ((player->curDepth)<(player->depth))
    {
      if (deepenLookaheadBy1(player,
                             &(player->curBoard),
                             player->lookahead)<0)
        {
          int tmp=errno;
          fprintf(stderr,"deepenLookahead(): could not prune tree: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }
      /*dumpTree(player->lookahead);*/
      player->curDepth++;
    }

  return 0;
}

static int weighTree(FanoronaPlayerLookaheadNode* tree,
                     int withSibs,
                     int isPlayer1)
{
  int treeWeight;
  if (tree->firstChild)
    treeWeight=weighTree(tree->firstChild,withSibs,isPlayer1);
  else
    treeWeight=tree->counts.mine-tree->counts.other;

  if (withSibs)
    {
      FanoronaPlayerLookaheadNode* cur;
      for (cur=tree->nextSib; cur; cur=cur->nextSib)
        {
          int curWeight=weighTree(cur,withSibs,isPlayer1);
          if (tree->move.byMe)
            {
              if (curWeight>treeWeight)
                treeWeight=curWeight;
            }
          else
            {
              if (curWeight<treeWeight)
                treeWeight=curWeight;
            }
        }
    }

  return treeWeight;
}

static int pickMove(FanoronaPlayer* player,
                    FanoronaPlayerMove* move)
{
  FanoronaPlayerLookaheadNode* cur;
  int bestWeight;
  FanoronaPlayerMove* bestMove=0;
  for (cur=player->lookahead; cur; cur=cur->nextSib)
    {
      int curWeight=weighTree(cur,0,player->isPlayer1);
      if ((!bestMove) || (curWeight>bestWeight))
        {
          bestMove=&(cur->move);
          bestWeight=curWeight;
        }
    }

  if (!bestMove)
    {
      fprintf(stderr,"pickMove(): can't find any moves\n");
      errno=ENOENT;
      return -1;
    }

  (*move)=(*bestMove);
  return 0;
}

int fanoronaPlayerMakeMove(FanoronaPlayer* player,
                           int* colFrom,
                           int* rowFrom,
                           int* colTo,
                           int* rowTo,
                           int* byApproach)
{
  FanoronaPlayerMove move;

  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!colFrom)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): colFrom is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!rowFrom)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): rowFrom is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!colTo)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): colTo is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!rowTo)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): rowTo is null\n");
      errno=EFAULT;
      return -1;
    }

  if (deepenLookahead(player)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMakeMove(): could not deepen tree: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  if (pickMove(player,&move)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMakeMove(): could not pick move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  /*printf("Picked move: ");
  dumpMove(&move);*/

  if (applyMove(player,&move)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMakeMove(): could not apply move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  (*colFrom)=move.from.col;
  (*rowFrom)=move.from.row;
  (*colTo)=move.to.col;
  (*rowTo)=move.to.row;
  (*byApproach)=move.byApproach;

  return 0;
}

int fanoronaPlayerMoveMade(FanoronaPlayer* player,
                           int colFrom,
                           int rowFrom,
                           int colTo,
                           int rowTo,
                           int byApproach)
{
  FanoronaPlayerMove moveMade;
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerMoveMade(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  moveMade.byMe=0;
  moveMade.from.col=colFrom;
  moveMade.from.row=rowFrom;
  moveMade.to.col=colTo;
  moveMade.to.row=rowTo;
  moveMade.byApproach=byApproach;

  if (applyMove(player,&moveMade)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMoveMade(): could not apply move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  return 0;
}
