/* Copyright (c) 2000 John Stracke <fanorona@thibault.org>
 #
 * This file is part of Fanorona.
 #
 * Fanorona is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version. 
 #
 * Fanorona is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details. 
 #
 * You should have received a copy of the GNU General Public License along
 * with program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <fanorona/player.h>

#include "fakeBoard.h"

typedef struct _FanoronaPlayerLookaheadNode FanoronaPlayerLookaheadNode;

typedef struct {
  int byMe;
  struct {
    int col,row;
  } from,to;
  int byApproach;
} FanoronaPlayerMove;

typedef struct {
  FanoronaPlayerMove move;
  int weight;
  int turnIsOver;
} PotentialMove;

typedef struct {
  int len;
  struct {
    int col,row;
  } pos[FANORONA_NUM_COLS*FANORONA_NUM_ROWS];
} Path;

struct _FanoronaPlayer {
  int isPlayer1,depth;
  FanoronaBoard curBoard;
  Path turnSoFar;
};

/* returns a random # from 1 to max */
static int rnd(int max)
{
  {
    static int inited=0;
    if (!inited)
      {
        srand(time(0));
        inited=1;
      }
  }

  return (rand()%max)+1;
}

static void pathInit(Path* path)
{
  path->len=0;
#if 1
  {
    int i,N;
    for (i=0, N=sizeof(path->pos)/sizeof(path->pos[0]); i<N; i++)
      path->pos[i].col=path->pos[i].row=-1;
  }
#endif
}

static int pathContains(Path* path,
                        int col,
                        int row)
{
  int i;
  for (i=0; i<(path->len); i++)
    {
      if ((path->pos[i].col)!=col)
        continue;
      if ((path->pos[i].row)!=row)
        continue;
      return 1;
    }

  return 0;
}

static int pathEndsWith(Path* path,
                        int col,
                        int row)
{
  int i=path->len-1;
  if (i<0)
    return 0;
  if ((path->pos[i].col)!=col)
    return 0;
  if ((path->pos[i].row)!=row)
    return 0;
  return 1;
}

static int pathAppend(Path* path,
                       int col,
                       int row)
{
  int i;
  if (pathEndsWith(path,col,row))
    return 0;

  i=(path->len)++;
  path->pos[i].col=col;
  path->pos[i].row=row;
  return 1;
}

static void indent(int cols)
{
  while (cols>0)
    {
      printf(" ");
      cols--;
    }
}

static void _dumpMove(const FanoronaPlayerMove* move,
                      int indents)
{
  indent(indents);
  printf("<move byMe='%s' from='(%d,%d)' to='(%d,%d)' by='%s'/>\n",
         (move->byMe ? "yes" : "no"),
         move->from.col,
         move->from.row,
         move->to.col,
         move->to.row,
         (move->byApproach ? "approach" : "withdrawal"));
  fflush(stdout); fflush(stderr);
}

static void dumpMove(const FanoronaPlayerMove* move)
{
  _dumpMove(move,0);
}

static int copyAndMove(FanoronaBoard* dest,
                       const FanoronaBoard* src,
                       const FanoronaPlayerMove* move)
{
  if (fanoronaBoardCopyAndMove(dest,
                               src,
                               move->from.col,
                               move->from.row,
                               move->to.col,
                               move->to.row,
                               move->byApproach,
                               0)
      )
    {
      printf("copyAndMove() failed: move is ");
      dumpMove(move);
      return -1;
    }

  return 0;
}

int fanoronaPlayerOtherPlayerMoveValid(FanoronaPlayer* player,
                                       int colFrom,
                                       int rowFrom,
                                       int colTo,
                                       int rowTo,
                                       int* canApproach,
                                       int* canWithdraw)
{
  FanoronaBoardCellState myState,otherState,movingState,nonMovingState;
  int _canApproach,_canWithdraw,dc,dr;

  if (!fanoronaBoardIsLegalPos(colTo,rowTo))
    return 0;

  if (fanoronaBoardGetCell(&(player->curBoard),colTo,rowTo)
      !=fanoronaBoardCellStateEmpty
      )
    return 0;

  if (!fanoronaBoardTestAdjacent(colFrom,rowFrom,colTo,rowTo))
    return 0;

  if (player->isPlayer1)
    {
      myState=fanoronaBoardCellStatePlayer1;
      otherState=fanoronaBoardCellStatePlayer2;
    }
  else
    {
      myState=fanoronaBoardCellStatePlayer2;
      otherState=fanoronaBoardCellStatePlayer1;
    }

  movingState=otherState;
  nonMovingState=myState;

  dc=colTo-colFrom;
  dr=rowTo-rowFrom;

  {
    if ((player->turnSoFar.len)>1)
      {
        int lastDc=( (player->turnSoFar.pos[player->turnSoFar.len-1].col)
                     -(player->turnSoFar.pos[player->turnSoFar.len-2].col)
                     );
        int lastDr=( (player->turnSoFar.pos[player->turnSoFar.len-1].row)
                     -(player->turnSoFar.pos[player->turnSoFar.len-2].row)
                     );

        if ((lastDc==dc) && (lastDr==dr))
          return 0;
      }
  }

  _canApproach
    =(   fanoronaBoardIsLegalPos(colTo+dc,rowTo+dr)
         && (fanoronaBoardGetCell(&(player->curBoard),colTo+dc,rowTo+dr)
             ==nonMovingState)
         );
  _canWithdraw
    =(   fanoronaBoardIsLegalPos(colFrom-dc,rowFrom-dr)
         && (fanoronaBoardGetCell(&(player->curBoard),colFrom-dc,rowFrom-dr)
             ==nonMovingState)
         );

  if (canApproach)
    *canApproach=_canApproach;
  if (canWithdraw)
    *canWithdraw=_canWithdraw;

  if (!(_canApproach || _canWithdraw))
    return (player->turnSoFar.len==0);

  if (pathContains(&(player->turnSoFar),colTo,rowTo))
    return 0;

  /*printf("From (%d,%d), can capture by %s to (%d,%d)\n",
         colFrom,rowFrom,
         (_canApproach ? (_canWithdraw
                          ? "approach or withdrawal"
                          : "approach"
                          )
          : "withdrawal"
          ),
         colTo,rowTo
         );*/
  return 1;
}

int fanoronaPlayerOtherPlayerCanCapture(FanoronaPlayer* player,
                                        int colFrom,
                                        int rowFrom)
{
  int colTo,rowTo;

  for (colTo=colFrom-1; colTo<=(colFrom+1); colTo++)
    for (rowTo=rowFrom-1; rowTo<=(rowFrom+1); rowTo++)
      {
        if (fanoronaPlayerOtherPlayerMoveValid(player,
                                               colFrom,rowFrom,
                                               colTo,rowTo,
                                               0,0))
          return 1;
      }

  return 0;
}

static int applyMove(FanoronaPlayer* player,
                     const FanoronaPlayerMove* move,
                     int* turnIsOver)
{
  int numCaptured;
  if (fanoronaBoardMove(&(player->curBoard),
                        move->from.col,
                        move->from.row,
                        move->to.col,
                        move->to.row,
                        move->byApproach,
                        &numCaptured)<0)
    {
      int tmp=errno;
      fprintf(stderr,"applyMove(): could not do move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  if (turnIsOver)
    {
      (*turnIsOver)=!(numCaptured
                      && fanoronaPlayerOtherPlayerCanCapture(player,
                                                             move->to.col,
                                                             move->to.row)
                      );
    }

  return 0;
}

int fanoronaPlayerInit(FanoronaPlayer* player,
                       int isPlayer1,
                       int depth)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerInit(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (fanoronaBoardInit(&(player->curBoard))<0)
    {
      fprintf(stderr,"fanoronaPlayerInit(): can't init board\n");
      errno=EFAULT;
      return -1;
    }

  if (fanoronaBoardSetup(&(player->curBoard))<0)
    {
      fprintf(stderr,"fanoronaPlayerInit(): can't set up board\n");
      errno=EFAULT;
      return -1;
    }

  player->isPlayer1=isPlayer1;
  player->depth=depth;

  pathInit(&(player->turnSoFar));

  return 0;
}

int fanoronaPlayerNewGame(FanoronaPlayer* player)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerInit(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (fanoronaBoardSetup(&(player->curBoard))<0)
    {
      fprintf(stderr,"fanoronaPlayerInit(): can't set up board\n");
      errno=EFAULT;
      return -1;
    }

  pathInit(&(player->turnSoFar));

  return 0;
}

FanoronaPlayer* fanoronaPlayerNew(int isPlayer1,
                                  int depth)
{
  FanoronaPlayer* res=(FanoronaPlayer*)malloc(sizeof(FanoronaPlayer));
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerNew(): can't malloc player: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  if (fanoronaPlayerInit(res,isPlayer1,depth)<0)
    {
      int tmp=errno;
      free(res);
      fprintf(stderr,"fanoronaPlayerNew(): can't init player: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  return res;
}

void fanoronaPlayerFree(FanoronaPlayer* player)
{
  if (!player)
    return;

  free(player);
}

int fanoronaPlayerSetDepth(FanoronaPlayer* player,
                           int depth)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerSetDepth(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (depth<=0)
    {
      fprintf(stderr,"fanoronaPlayerSetDepth(%d): depth must be positive\n",
              depth);
      errno=EINVAL;
      return -1;
    }

  player->depth=depth;
  return 0;
}

FanoronaBoard* fanoronaPlayerGetBoard(FanoronaPlayer* player)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerGetBoard(): player is null\n");
      errno=EFAULT;
      return 0;
    }
  
  return &(player->curBoard);
}

int fanoronaPlayerSetBoard(FanoronaPlayer* player,
                           const FanoronaBoard* board)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerSetBoard(): player is null\n");
      errno=EFAULT;
      return -1;
    }
  
  if (!board)
    {
      fprintf(stderr,"fanoronaPlayerSetBoard(): board is null\n");
      errno=EFAULT;
      return -1;
    }

  return fanoronaBoardCopy(&(player->curBoard),board);
}

int fanoronaPlayerIsPlayer1(FanoronaPlayer* player)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerGetBoard(): player is null\n");
      errno=EFAULT;
      return 0;
    }
  
  return player->isPlayer1;
}

static int pickNextMove(const FanoronaPlayer* player,
                        int depthLeft,
                        const FanoronaBoard* board,
                        int isMyMove,
                        int mustCaptureWithCol,
                        int mustCaptureWithRow,
                        FanoronaPlayerMove* move,
                        int* weight,
                        int* turnIsOver,
                        Path* path)
{
  FanoronaBoardCellState myState,otherState,movingState,nonMovingState;
  PotentialMove potentialMoves[FANORONA_NUM_COLS*FANORONA_NUM_ROWS*8*2];
  int numPotentialMoves=0;

  {
    int numForPlayer1,numForPlayer2,numMine,numOther;

    fanoronaBoardCountSides(board,
                            &numForPlayer1,
                            &numForPlayer2);

    if (player->isPlayer1)
      {
        numMine=numForPlayer1;
        numOther=numForPlayer2;
      }
    else
      {
        numMine=numForPlayer2;
        numOther=numForPlayer1;
      }

    if ((numForPlayer1==0) || (numForPlayer2==0))
      {
        if (weight)
          (*weight)=numMine-numOther;
        if (turnIsOver)
          (*turnIsOver)=0;
        return 0;
      }
  }

  if (player->isPlayer1)
    {
      myState=fanoronaBoardCellStatePlayer1;
      otherState=fanoronaBoardCellStatePlayer2;
    }
  else
    {
      myState=fanoronaBoardCellStatePlayer2;
      otherState=fanoronaBoardCellStatePlayer1;
    }

  if (isMyMove)
    {
      movingState=myState;
      nonMovingState=otherState;
    }
  else
    {
      movingState=otherState;
      nonMovingState=myState;
    }

  {
    int colFrom,rowFrom;
    int firstCol,lastCol,firstRow,lastRow;

    if (mustCaptureWithCol>=0)
      {
        firstCol=lastCol=mustCaptureWithCol;
        firstRow=lastRow=mustCaptureWithRow;
      }
    else
      {
        firstCol=firstRow=0;
        lastRow=FANORONA_NUM_ROWS-1;
        lastCol=FANORONA_NUM_COLS-1;
      }

    for (colFrom=firstCol; colFrom<=lastCol; colFrom++)
      for (rowFrom=firstRow; rowFrom<=lastRow; rowFrom++)
        {
          int colTo,rowTo;
          if (fanoronaBoardGetCell(board,colFrom,rowFrom)!=movingState)
            continue;

          for (colTo=colFrom-1; colTo<=(colFrom+1); colTo++)
            for (rowTo=rowFrom-1; rowTo<=(rowFrom+1); rowTo++)
              {
                int byApproach,canApproach,canWithdraw,dc,dr;

                if ((colTo==colFrom) && (rowTo==rowFrom))
                  continue;

                if (!fanoronaBoardIsLegalPos(colTo,rowTo))
                  continue;
                
                if (fanoronaBoardGetCell(board,colTo,rowTo)
                    !=fanoronaBoardCellStateEmpty
                    )
                  continue;

                if (!fanoronaBoardTestAdjacent(colFrom,rowFrom,colTo,rowTo))
                  continue;

                if (pathContains(path,colTo,rowTo))
                  continue;

                dc=colTo-colFrom;
                dr=rowTo-rowFrom;

                if ((path->len)>1)
                  {
                    int lastDc=( (path->pos[path->len-1].col)
                                -(path->pos[path->len-2].col)
                                 );
                    int lastDr=( (path->pos[path->len-1].row)
                                -(path->pos[path->len-2].row)
                                 );

                    /*printf("(lastDc,lastDr)==(%d,%d); (dc,dr)==(%d,%d)\n",
                           lastDc,lastDr,dc,dr);*/

                    if ((lastDc==dc) && (lastDr==dr))
                      continue;
                  }

                canApproach
                  =(   fanoronaBoardIsLegalPos(colTo+dc,rowTo+dr)
                    && (fanoronaBoardGetCell(board,colTo+dc,rowTo+dr)
                        ==nonMovingState)
                       );
                canWithdraw
                  =(   fanoronaBoardIsLegalPos(colFrom-dc,rowFrom-dr)
                    && (fanoronaBoardGetCell(board,colFrom-dc,rowFrom-dr)
                        ==nonMovingState)
                       );

                if (!(canApproach || canWithdraw))
                  canApproach=1;

                for (byApproach=0; byApproach<2; byApproach++)
                  {
                    FanoronaBoard potentialBoard;
                    PotentialMove* potentialMove=0;
                    int isCapture;

                    if (byApproach)
                      {
                        if (!canApproach)
                          continue;
                      }
                    else
                      {
                        if (!canWithdraw)
                          continue;
                      }

                    potentialMove=&(potentialMoves[numPotentialMoves++]);

                    potentialMove->move.byMe=isMyMove;
                    potentialMove->move.from.col=colFrom;
                    potentialMove->move.from.row=rowFrom;
                    potentialMove->move.to.col=colTo;
                    potentialMove->move.to.row=rowTo;
                    potentialMove->move.byApproach=byApproach;
                    
                    /*printf("Trying move: from (%d,%d) to (%d,%d), by %s\n",
                           colFrom,rowFrom,colTo,rowTo,
                           (byApproach ? "approach" : "withdrawal"));*/

                    if (fanoronaBoardCopyAndMove(&potentialBoard,board,
                                                 colFrom,rowFrom,
                                                 colTo,rowTo,
                                                 byApproach,
                                                 &isCapture)<0)
                      {
                        int tmp=errno;
                        fprintf(stderr,
                                "pickNextMove(): can't apply possible move: %s\n",
                                strerror(tmp));
                        tmp=errno;
                        return -1;
                      }

                    if ((mustCaptureWithRow>=0) && (!isCapture))
                      {
                        numPotentialMoves--;
                        continue;
                      }

                    if (depthLeft==0)
                      {
                        int numForPlayer1,numForPlayer2;
                        int numMine,numOther;
                        fanoronaBoardCountSides(&potentialBoard,
                                                &numForPlayer1,
                                                &numForPlayer2);
                        if (player->isPlayer1)
                          {
                            numMine=numForPlayer1;
                            numOther=numForPlayer2;
                          }
                        else
                          {
                            numMine=numForPlayer2;
                            numOther=numForPlayer1;
                          }
                        potentialMove->weight=numMine-numOther;
                        potentialMove->turnIsOver=(   (!isCapture)
                                                   || (numMine<=0)
                                                   || (numOther<=0)
                                                      );
                      }
                    else
                      {
                        int thisTurnIsOver=(!isCapture);
                        if (!thisTurnIsOver)
                          {
                            int numAppended=0;
                            numAppended+=pathAppend(path,colFrom,rowFrom);
                            numAppended+=pathAppend(path,colTo,rowTo);
                            if (pickNextMove(player,
                                             depthLeft-1,
                                             &potentialBoard,
                                             isMyMove,
                                             colTo,rowTo,
                                             0,
                                             &(potentialMove->weight),
                                             0,
                                             path)
                                <0)
                              {
                               /* if (depthLeft==(player->depth))
                                  printf("(%d,%d)->(%d,%d): Captured, but over\n",
                                         colFrom,rowFrom,
                                         colTo,rowTo);*/
                                thisTurnIsOver=1;
                              }
                            path->len-=numAppended;
                          }

                        potentialMove->turnIsOver=thisTurnIsOver;

                        if (thisTurnIsOver)
                          {
                            Path otherPath;
                            pathInit(&otherPath);
                            {
                              int numForPlayer1,numForPlayer2;
                              int numMine,numOther;
                              int numNextMove;

                              fanoronaBoardCountSides(&potentialBoard,
                                                      &numForPlayer1,
                                                      &numForPlayer2);
                              if (player->isPlayer1)
                                {
                                  numMine=numForPlayer1;
                                  numOther=numForPlayer2;
                                }
                              else
                                {
                                  numMine=numForPlayer2;
                                  numOther=numForPlayer1;
                                }

                              if (!isMyMove)
                                numNextMove=numMine;
                              else
                                numNextMove=numOther;

                              if (numNextMove<=0)
                                {
                                  potentialMove->weight=numMine-numOther;
                                  return 0;
                                }
                            }

                            if (pickNextMove(player,
                                             depthLeft-1,
                                             &potentialBoard,
                                             !isMyMove,
                                             -1,-1,
                                             0,
                                             &(potentialMove->weight),
                                             0,
                                             &otherPath)
                                <0)
                              return -1;
                          }
                      }
                  }
              }
        }
  }

  if (numPotentialMoves<=0)
    {
      if (depthLeft==(player->depth))
        fprintf(stderr,"pickNextMove(): found no moves\n");
      errno=ENOENT;
      return -1;
    }

  {
    int i,found=0,numFoundWithThisWeight=1;
    int minmax=potentialMoves[0].weight;
    for (i=1; i<numPotentialMoves; i++)
      {
        int curWeight=potentialMoves[i].weight;
        if (curWeight==minmax)
          {
            numFoundWithThisWeight++;
            if (rnd(numFoundWithThisWeight)==1)
              found=i;
            continue;
          }

        if (isMyMove)
          {
            if (curWeight>minmax)
              {
                minmax=curWeight;
                found=i;
              }
          }
        else
          {
            if (curWeight<minmax)
              {
                minmax=curWeight;
                found=i;
              }
          }
      }

    if (move)
      (*move)=potentialMoves[found].move;
    if (weight)
      (*weight)=minmax;
    if (turnIsOver)
      (*turnIsOver)=potentialMoves[found].turnIsOver;
    return 0;
  }
}

static int pickMove(const FanoronaPlayer* player,
                    FanoronaPlayerMove* move,
                    int* turnIsOver)
{
  int mustCaptureWithCol,mustCaptureWithRow,N;
  Path path;

  {
    int numForPlayer1,numForPlayer2;

    fanoronaBoardCountSides(&(player->curBoard),
                            &numForPlayer1,
                            &numForPlayer2);
    if ((numForPlayer1==1) || (numForPlayer2==1))
      {
        printf("Only one left\n");
      }
  }

  memcpy(&path,&(player->turnSoFar),sizeof(path));

  if (N=(player->turnSoFar.len))
    {
      mustCaptureWithCol
        =player->turnSoFar.pos[N-1].col;
      mustCaptureWithRow
        =player->turnSoFar.pos[N-1].row;
    }
  else
    {
      mustCaptureWithCol=-1;
      mustCaptureWithRow=-1;
    }

  return pickNextMove(player,player->depth,&(player->curBoard),1,
                      mustCaptureWithCol,
                      mustCaptureWithRow,
                      move,0,
                      turnIsOver,
                      &path);
}

int fanoronaPlayerMakeMove(FanoronaPlayer* player,
                           int* colFrom,
                           int* rowFrom,
                           int* colTo,
                           int* rowTo,
                           int* byApproach,
                           int* turnIsOver)
{
  FanoronaPlayerMove move;

  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!colFrom)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): colFrom is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!rowFrom)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): rowFrom is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!colTo)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): colTo is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!rowTo)
    {
      fprintf(stderr,"fanoronaPlayerMakeMove(): rowTo is null\n");
      errno=EFAULT;
      return -1;
    }

  if (pickMove(player,&move,turnIsOver)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMakeMove(): could not pick move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  /*printf("Picked move: ");
  dumpMove(&move);
  fanoronaBoardPrint(stdout,&(player->curBoard));*/

  if (applyMove(player,&move,0)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMakeMove(): could not apply move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  if (!(*turnIsOver))
    {
      int numForPlayer1,numForPlayer2;

      fanoronaBoardCountSides(&(player->curBoard),
                              &numForPlayer1,
                              &numForPlayer2);
      if ((!numForPlayer1) || (!numForPlayer2))
        (*turnIsOver)=1;
    }

  if (*turnIsOver)
    {
      pathInit(&(player->turnSoFar));
    }
  else
    {
      pathAppend(&(player->turnSoFar),move.from.col,move.from.row);
      pathAppend(&(player->turnSoFar),move.to.col,move.to.row);
    }

  (*colFrom)=move.from.col;
  (*rowFrom)=move.from.row;
  (*colTo)=move.to.col;
  (*rowTo)=move.to.row;
  (*byApproach)=move.byApproach;

  return 0;
}

int fanoronaPlayerMoveMade(FanoronaPlayer* player,
                           int colFrom,
                           int rowFrom,
                           int colTo,
                           int rowTo,
                           int byApproach,
                           int* turnIsOver)
{
  FanoronaPlayerMove moveMade;
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerMoveMade(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  {
    int N;
    if (N=(player->turnSoFar.len))
      {
        {
          int mustCaptureWithCol,mustCaptureWithRow;
          mustCaptureWithCol
            =player->turnSoFar.pos[N-1].col;
          mustCaptureWithRow
            =player->turnSoFar.pos[N-1].row;

          if (   ((mustCaptureWithCol)!=colFrom)
                 || ((mustCaptureWithRow)!=rowFrom)
                 )
            {
              fprintf(stderr,
                      "fanoronaPlayerMoveMade(): all moves in turn must be with same piece\n");
              errno=EINVAL;
              return -1;
            }
        }
        {
          int i;
          for (i=0; i<(N-1); i++)
            {
              if ((player->turnSoFar.pos[i].col)!=colTo)
                continue;
              if ((player->turnSoFar.pos[i].row)!=rowTo)
                continue;
              fprintf(stderr,
                      "fanoronaPlayerMoveMade(): may not cross back over own path\n");
              errno=EINVAL;
              return -1;
            }
        }
      }
  }

  moveMade.byMe=0;
  moveMade.from.col=colFrom;
  moveMade.from.row=rowFrom;
  moveMade.to.col=colTo;
  moveMade.to.row=rowTo;
  moveMade.byApproach=byApproach;

  pathAppend(&(player->turnSoFar),colFrom,rowFrom);
  pathAppend(&(player->turnSoFar),colTo,rowTo);

  if (applyMove(player,&moveMade,turnIsOver)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMoveMade(): could not apply move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  if (*turnIsOver)
    pathInit(&(player->turnSoFar));

  return 0;
}

int fanoronaPlayerOtherPlayerResetPath(FanoronaPlayer* player)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerOtherPlayerResetPath(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  pathInit(&(player->turnSoFar));
  return 0;
}

void* fanoronaPlayerGetState(FanoronaPlayer* player,
                             int* len)
{
  Path* res;

  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerGetState(): player is null\n");
      errno=EFAULT;
      return 0;
    }

  if (!len)
    {
      fprintf(stderr,"fanoronaPlayerGetState(): len is null\n");
      errno=EFAULT;
      return 0;
    }

  res=(Path*)malloc(sizeof(Path));
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerGetState(): can't malloc() res: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  memcpy(res,&(player->turnSoFar),sizeof(Path));

  (*len)=sizeof(Path);
  return res;
}

int fanoronaPlayerRestoreState(FanoronaPlayer* player,
                               const void* data,
                               int len)
{
  if (!player)
    {
      fprintf(stderr,"fanoronaPlayerRestoreState(): player is null\n");
      errno=EFAULT;
      return -1;
    }

  if (!data)
    {
      fprintf(stderr,"fanoronaPlayerRestoreState(): data is null\n");
      errno=EFAULT;
      return -1;
    }

  if (len!=sizeof(Path))
    {
      fprintf(stderr,
              "fanoronaPlayerRestoreState(): len is wrong (should be %d)\n",
              sizeof(Path));
      errno=EINVAL;
      return -1;
    }

  memcpy(&(player->turnSoFar),data,sizeof(Path));

  return 0;
}
